/*
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of this code is David Baum.
 * Portions created by David Baum are Copyright (C) 1998 David Baum.
 * All Rights Reserved.
 */

#include "RCX_SpyboticsLinker.h"
#include "RCX_Image.h"
#include "RCX_Target.h"

RCX_SpyboticsLinker::RCX_SpyboticsLinker()
{
}


void RCX_SpyboticsLinker::Generate(const RCX_Image &image, std::vector<UByte> &output)
{
	fTocOffset = 0;
	fContentOffset = kRCX_SpyboticsTocSize * 2;
	fOutput = &output;
	fOutput->reserve(kRCX_SpyboticsTocSize*2 + 1000);
	fOutput->resize(fContentOffset);
	
	ExtractCode(image);

	GenerateSubs();
	GenerateTasks();
	GenerateSounds();
	GenerateAnimations();
	
	// data area
	AddEntry();
	
	// end marker
	AddEntry();


	// dump the binary image for debugging
/*
	int length = fOutput->size();
	const UByte* data = &(*fOutput)[0];
	for(int i=0; i<length; ++i) {
		if ((i%16)==0) printf("\n%04x: ", i);
		printf("%02x ", data[i]);
	}
	printf("\n");
/**/

}



void RCX_SpyboticsLinker::GenerateSubs()
{
	// subroutines
	UByte endSub = 0xf6;
	for(int i=0; i<kRCX_SpyboticsMaxSubs; ++i) {
		const RCX_Image::Fragment *f = fSubs[i];
		
		AddEntry();
		if (f) {
			AddData(f->GetData(), f->GetLength());
		}
		AddData(&endSub, 1);
	}
}


void RCX_SpyboticsLinker::GenerateTasks()
{
	// subroutines
	UByte endTask[2] = {0x81, 0};

	for(int i=0; i<kRCX_SpyboticsMaxTasks; ++i) {
		const RCX_Image::Fragment *f = fTasks[i];
		
		AddEntry();
		if (f) {
			AddData(f->GetData(), f->GetLength());
		}
		endTask[1] = i+8;
		AddData(endTask, 2);
	}
}


void RCX_SpyboticsLinker::GenerateSounds()
{
	// sounds
	UByte emptySound = 0;
	for(int i=0; i<kRCX_SpyboticsMaxSounds; ++i) {
		AddEntry(&emptySound, 1);
	}
}


void RCX_SpyboticsLinker::GenerateAnimations()
{
	// animations
	UByte emptyAnimation[2] = {0, 0};
	for(int i=0; i<kRCX_SpyboticsMaxAnimations; ++i) {
		AddEntry(emptyAnimation, 2);
	}
}



void RCX_SpyboticsLinker::ExtractCode(const RCX_Image &image)
{
	int i;
	
	for(i=0; i<kRCX_SpyboticsMaxTasks; ++i) fTasks[i] = 0;
	for(i=0; i<kRCX_SpyboticsMaxSubs; ++i) fSubs[i] = 0;
	
	for(i=0; i<image.GetFragmentCount(); ++i)
	{
		const RCX_Image::Fragment &f = image.GetFragment(i);
		
		if (f.GetType() == kRCX_TaskFragment) 
			fTasks[f.GetNumber() - 8] = &f;
		else
			fSubs[f.GetNumber() - 128] = &f;
	}	
}



void RCX_SpyboticsLinker::AddEntry(const UByte *data, int length)
{
	// add TOC entry
	int address = fContentOffset + 0x8100;
	(*fOutput)[fTocOffset++] = (address) & 0xff;
	(*fOutput)[fTocOffset++] = (address >> 8) & 0xff;
	
	if (length) AddData(data, length);
}


void RCX_SpyboticsLinker::AddData(const UByte *data, int length)
{
	if (!length) return;
	
	// add data
	fOutput->resize(fContentOffset + length);
	memcpy(&(*fOutput)[fContentOffset], data, length);
	fContentOffset += length;
}
